/*
 * $Id: reboot-proc.c,v 1.1 2007/05/23 14:06:28 clucas Exp clucas $
 * 
 *  Copyright (c) 2007  Christophe Lucas <clucas@rotomalug.org>
 * 
 *  reboot-proc:  Allow to emergency reboot via write to /proc entry
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/tty.h>
#include <linux/mount.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/reboot.h>
#include <linux/sysrq.h>
#include <linux/kbd_kern.h>
#include <linux/quotaops.h>
#include <linux/smp_lock.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/suspend.h>
#include <linux/writeback.h>
#include <linux/buffer_head.h>          /* for fsync_bdev() */
#include <linux/swap.h>
#include <linux/spinlock.h>
#include <linux/vt_kern.h>
#include <linux/workqueue.h>
#include <linux/kexec.h>
#include <linux/irq.h>
#include <linux/lockdep.h>

MODULE_AUTHOR("Christophe Lucas <clucas@rotomalug.org>");
MODULE_VERSION("v0.1");
MODULE_DESCRIPTION("Reboot from proc");
MODULE_SUPPORTED_DEVICE("none");
MODULE_LICENSE("GPL");

#define PROCFS_FILE_MAX_SIZE        1024

static char module_buffer[PROCFS_FILE_MAX_SIZE];

static unsigned long module_buffer_size = 0;

static void send_sig_all(int sig)
{
	struct task_struct *p;
	 
	for_each_process(p) {
		if (p->mm && p->pid != 1)
			/* Not swapper, init nor kernel thread */
			force_sig(sig, p);
	}
}

static int 
reboot_read_proc(char *page, char **start, off_t off, int count, int *eof,
		 void* data)
{
	if (off > 0) {
		return(0);
	} else {
		memcpy(page, module_buffer, module_buffer_size);
		return(module_buffer_size);
	}
}

static int 
reboot_write_proc(struct file *file, const char __user *buffer,
	unsigned long count, void *data)
{
	int result;

	module_buffer_size = count;
	if (module_buffer_size > PROCFS_FILE_MAX_SIZE ) {
		module_buffer_size = PROCFS_FILE_MAX_SIZE;
	}
 
	if (copy_from_user(module_buffer, buffer, module_buffer_size)) {
		result = -EFAULT;
        } else {
		module_buffer[count-1] = 0;
	}
	
	if (strncmp(module_buffer, "reboot", count) == 0) {
		printk(KERN_INFO "reboot-proc: we will now reboot...\n");
		local_irq_enable();
		send_sig_all(SIGKILL);
		emergency_restart();
	} 
	return module_buffer_size;
}

static int __init reboot_proc_module_init(void)
{
	struct proc_dir_entry *p;
	printk(KERN_INFO "reboot-proc: installed.\n");
	p = create_proc_read_entry("reboot-proc",  S_IFREG | S_IRUGO | S_IWUSR,
				NULL,
				(read_proc_t *)reboot_read_proc, NULL);
	if (p == NULL) {
		printk(KERN_ERR "reboot-proc: Unable to create proc entry.\n");
		return -EINVAL;
	} else {
		p->owner = THIS_MODULE;
		p->read_proc = (read_proc_t*) reboot_read_proc;
		p->write_proc = (write_proc_t*) reboot_write_proc;
		p->uid = 0;
		p->gid = 0;
		p->size = PROCFS_FILE_MAX_SIZE;
	}
	return 0;
}

static void __exit reboot_proc_module_cleanup(void)
{
	printk(KERN_INFO "reboot-proc: uninstalled.\n");
	remove_proc_entry("reboot-proc", NULL);
}

module_init(reboot_proc_module_init);
module_exit(reboot_proc_module_cleanup);

